@@ -3,7 +3,7 @@
  */
 
 /*
- * Copyright (C) 1986, 1988, 1989, 1991-2017 the Free Software Foundation, Inc.
+ * Copyright (C) 1986, 1988, 1989, 1991-2021 the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -43,21 +43,16 @@
 #include <config.h>
 #endif
 
-#ifndef _GNU_SOURCE
-#define _GNU_SOURCE	1	/* enable GNU extensions */
-#endif /* _GNU_SOURCE */
-
 #if defined(tandem_for_real) && ! defined(_SCO_DS)
 #define _XOPEN_SOURCE_EXTENDED 1
 #endif
 
 #include <stdio.h>
 #include <assert.h>
-#ifdef HAVE_LIMITS_H
 #include <limits.h>
-#endif /* HAVE_LIMITS_H */
 #include <ctype.h>
 #include <setjmp.h>
+#include <math.h>
 
 #include "gettext.h"
 #define _(msgid)  gettext(msgid)
@@ -69,15 +64,14 @@
 #endif /* LOCALEDIR */
 #endif
 
+#if !defined(__SUNPRO_C)
 #if !defined(__STDC__) || __STDC__ < 1
 #error "gawk no longer supports non-C89 environments (no __STDC__ or __STDC__ < 1)"
 #endif
-
-#if defined(HAVE_STDARG_H)
-#include <stdarg.h>
-#else
-#error "gawk no longer supports <varargs.h>. Please update your compiler and runtime"
 #endif
+
+#include <stdarg.h>
+#include <stdbool.h>
 #include <signal.h>
 #include <time.h>
 #include <errno.h>
@@ -89,18 +83,11 @@ extern int errno;
 #include <stdlib.h>
 #endif	/* not STDC_HEADERS */
 
-#ifdef HAVE_STDBOOL_H
-#include <stdbool.h>
-#else
-#include "missing_d/gawkbool.h"
-#endif
 
 /* We can handle multibyte strings.  */
 #include <wchar.h>
 #include <wctype.h>
 
-#include "mbsupport.h" /* defines stuff for DJGPP to fake MBS */
-
 #ifdef STDC_HEADERS
 #include <float.h>
 #endif
@@ -233,7 +220,7 @@ extern double gawk_strtod();
 #endif
 
 #if __GNUC__ < 2 || (__GNUC__ == 2 && __GNUC_MINOR__ < 7)
-# define __attribute__(x)
+# define __attribute__(arg)
 #endif
 
 #ifndef ATTRIBUTE_UNUSED
@@ -308,6 +295,12 @@ typedef union bucket_item {
 	} hi;
 } BUCKET;
 
+enum commenttype {
+	EOL_COMMENT = 1,
+	BLOCK_COMMENT,
+	FOR_COMMENT	// special case
+};
+
 /* string hash table */
 #define ahnext		hs.next
 #define	ahname		hs.name	/* a string index node */
@@ -326,6 +319,19 @@ struct exp_instruction;
 
 typedef int (*Func_print)(FILE *, const char *, ...);
 typedef struct exp_node **(*afunc_t)(struct exp_node *, struct exp_node *);
+typedef struct {
+	const char *name;
+	afunc_t init;
+	afunc_t type_of;	/* avoid reserved word typeof */
+	afunc_t lookup;
+	afunc_t exists;
+	afunc_t clear;
+	afunc_t remove;
+	afunc_t list;
+	afunc_t copy;
+	afunc_t dump;
+	afunc_t store;
+} array_funcs_t;
 
 /*
  * NOTE - this struct is a rather kludgey -- it is packed to minimize
@@ -338,7 +344,7 @@ typedef struct exp_node {
 				struct exp_node *lptr;
 				struct exp_instruction *li;
 				long ll;
-				afunc_t *lp;
+				const array_funcs_t *lp;
 			} l;
 			union {
 				struct exp_node *rptr;
@@ -352,14 +358,16 @@ typedef struct exp_node {
 				struct exp_node *extra;
 				void (*aptr)(void);
 				long xl;
+				void *cmnt;	// used by pretty printer
 			} x;
 			char *name;
 			size_t reserved;
 			struct exp_node *rn;
 			unsigned long cnt;
-			unsigned long reflags;
-#				define	CONSTANT	1
-#				define	FS_DFLT		2
+			enum reflagvals {
+				CONSTANT = 1,
+				FS_DFLT  = 2,
+			} reflags;
 		} nodep;
 
 		struct {
@@ -369,92 +377,94 @@ typedef struct exp_node {
 				mpfr_t mpnum;
 				mpz_t mpi;
 			} nm;
+			int rndmode;
 #else
 			AWKNUM fltnum;
 #endif
 			char *sp;
 			size_t slen;
-			long sref;
 			int idx;
 			wchar_t *wsp;
 			size_t wslen;
 			struct exp_node *typre;
+			enum commenttype comtype;
 		} val;
 	} sub;
 	NODETYPE type;
-	unsigned int flags;
+	enum flagvals {
+	/* type = Node_val */
+		/*
+		 * STRING and NUMBER are mutually exclusive, except for the special
+		 * case of an uninitialized value, represented internally by
+		 * Nnull_string. They represent the type of a value as assigned.
+		 * Nnull_string has both STRING and NUMBER attributes, but all other
+		 * scalar values should have precisely one of these bits set.
+		 *
+		 * STRCUR and NUMCUR are not mutually exclusive. They represent that
+		 * the particular type of value is up to date.  For example,
+		 *
+		 * 	a = 5		# NUMBER | NUMCUR
+		 * 	b = a ""	# Adds STRCUR to a, since a string value
+		 * 			# is now available. But the type hasn't changed!
+		 *
+		 * 	a = "42"	# STRING | STRCUR
+		 * 	b = a + 0	# Adds NUMCUR to a, since numeric value
+		 * 			# is now available. But the type hasn't changed!
+		 *
+		 * USER_INPUT is the joker.  When STRING|USER_INPUT is set, it means
+		 * "this is string data, but the user may have really wanted it to be a
+		 * number. If we have to guess, like in a comparison, turn it into a
+		 * number if the string is indeed numeric."
+		 * For example,    gawk -v a=42 ....
+		 * Here, `a' gets STRING|STRCUR|USER_INPUT and then when used where
+		 * a number is needed, it gets turned into a NUMBER and STRING
+		 * is cleared. In that case, we leave the USER_INPUT in place, so
+		 * the combination NUMBER|USER_INPUT means it is a strnum a.k.a. a
+		 * "numeric string".
+		 *
+		 * WSTRCUR is for efficiency. If in a multibyte locale, and we
+		 * need to do something character based (substr, length, etc.)
+		 * we create the corresponding wide character string and store it,
+		 * and add WSTRCUR to the flags so that we don't have to do the
+		 * conversion more than once.
+		 *
+		 * The NUMINT flag may be used with a value of any type -- NUMBER,
+		 * STRING, or STRNUM. It indicates that the string representation
+		 * equals the result of sprintf("%ld", <numeric value>). So, for
+		 * example, NUMINT should NOT be set if it's a strnum or string value
+		 * where the string is " 1" or "01" or "+1" or "1.0" or "0.1E1". This
+		 * is a hint to indicate that an integer array optimization may be
+		 * used when this value appears as a subscript.
+		 *
+		 * We hope that the rest of the flags are self-explanatory. :-)
+		 */
+		MALLOC	= 0x0001,       /* stptr can be free'd, i.e. not a field node pointing into a shared buffer */
+		STRING	= 0x0002,       /* assigned as string */
+		STRCUR	= 0x0004,       /* string value is current */
+		NUMCUR	= 0x0008,       /* numeric value is current */
+		NUMBER	= 0x0010,       /* assigned as number */
+		USER_INPUT = 0x0020,    /* user input: if NUMERIC then
+					 * a NUMBER */
+		INTLSTR	= 0x0040,       /* use localized version */
+		NUMINT	= 0x0080,       /* numeric value is an integer */
+		INTIND	= 0x0100,	/* integral value is array index;
+					 * lazy conversion to string.
+					 */
+		WSTRCUR	= 0x0200,       /* wide str value is current */
+		MPFN	= 0x0400,       /* arbitrary-precision floating-point number */
+		MPZN	= 0x0800,       /* arbitrary-precision integer */
+		NO_EXT_SET = 0x1000,    /* extension cannot set a value for this variable */
+		NULL_FIELD = 0x2000,    /* this is the null field */
 
-/* type = Node_val */
-	/*
-	 * STRING and NUMBER are mutually exclusive, except for the special
-	 * case of an uninitialized value, represented internally by
-	 * Nnull_string. They represent the type of a value as assigned.
-	 * Nnull_string has both STRING and NUMBER attributes, but all other
-	 * scalar values should have precisely one of these bits set.
-	 *
-	 * STRCUR and NUMCUR are not mutually exclusive. They represent that
-	 * the particular type of value is up to date.  For example,
-	 *
-	 * 	a = 5		# NUMBER | NUMCUR
-	 * 	b = a ""	# Adds STRCUR to a, since a string value
-	 * 			# is now available. But the type hasn't changed!
-	 *
-	 * 	a = "42"	# STRING | STRCUR
-	 * 	b = a + 0	# Adds NUMCUR to a, since numeric value
-	 * 			# is now available. But the type hasn't changed!
-	 *
-	 * USER_INPUT is the joker.  When STRING|USER_INPUT is set, it means
-	 * "this is string data, but the user may have really wanted it to be a
-	 * number. If we have to guess, like in a comparison, turn it into a
-	 * number if the string is indeed numeric."
-	 * For example,    gawk -v a=42 ....
-	 * Here, `a' gets STRING|STRCUR|USER_INPUT and then when used where
-	 * a number is needed, it gets turned into a NUMBER and STRING
-	 * is cleared. In that case, we leave the USER_INPUT in place, so
-	 * the combination NUMBER|USER_INPUT means it is a strnum a.k.a. a
-	 * "numeric string".
-	 *
-	 * WSTRCUR is for efficiency. If in a multibyte locale, and we
-	 * need to do something character based (substr, length, etc.)
-	 * we create the corresponding wide character string and store it,
-	 * and add WSTRCUR to the flags so that we don't have to do the
-	 * conversion more than once.
-	 *
-	 * The NUMINT flag may be used with a value of any type -- NUMBER,
-	 * STRING, or STRNUM. It indicates that the string representation
-	 * equals the result of sprintf("%ld", <numeric value>). So, for
-	 * example, NUMINT should NOT be set if it's a strnum or string value
-	 * where the string is " 1" or "01" or "+1" or "1.0" or "0.1E1". This
-	 * is a hint to indicate that an integer array optimization may be
-	 * used when this value appears as a subscript.
-	 *
-	 * We hope that the rest of the flags are self-explanatory. :-)
-	 */
-#		define	MALLOC	0x0001       /* stptr can be free'd, i.e. not a field node pointing into a shared buffer */
-#		define	STRING	0x0002       /* assigned as string */
-#		define	STRCUR	0x0004       /* string value is current */
-#		define	NUMCUR	0x0008       /* numeric value is current */
-#		define	NUMBER	0x0010       /* assigned as number */
-#		define	USER_INPUT 0x0020    /* user input: if NUMERIC then
-		                              * a NUMBER */
-#		define	INTLSTR	0x0040       /* use localized version */
-#		define	NUMINT	0x0080       /* numeric value is an integer */
-#		define	INTIND	0x0100	     /* integral value is array index;
-		                              * lazy conversion to string.
-		                              */
-#		define	WSTRCUR	0x0200       /* wide str value is current */
-#		define	MPFN	0x0400       /* arbitrary-precision floating-point number */
-#		define	MPZN	0x0800       /* arbitrary-precision integer */
-#		define	NO_EXT_SET 0x1000    /* extension cannot set a value for this variable */
-#		define	NULL_FIELD 0x2000    /* this is the null field */
-
-/* type = Node_var_array */
-#		define	ARRAYMAXED	0x4000       /* array is at max size */
-#		define	HALFHAT		0x8000       /* half-capacity Hashed Array Tree;
-		                                      * See cint_array.c */
-#		define	XARRAY		0x10000
-#		define	NUMCONSTSTR	0x20000	/* have string value for numeric constant */
-#		define  REGEX           0x40000 /* this is a typed regex */
+	/* type = Node_var_array */
+		ARRAYMAXED	= 0x4000,       /* array is at max size */
+		HALFHAT		= 0x8000,       /* half-capacity Hashed Array Tree;
+						 * See cint_array.c */
+		XARRAY		= 0x10000,
+		NUMCONSTSTR	= 0x20000,	/* have string value for numeric constant */
+		REGEX           = 0x40000,	/* this is a typed regex */
+	} flags;
+	long valref;
 } NODE;
 
 #define vname sub.nodep.name
@@ -491,8 +501,8 @@ typedef struct exp_node {
  */
 #define stptr	sub.val.sp
 #define stlen	sub.val.slen
-#define valref	sub.val.sref
 #define stfmt	sub.val.idx
+#define strndmode sub.val.rndmode
 #define wstptr	sub.val.wsp
 #define wstlen	sub.val.wslen
 #ifdef HAVE_MPFR
@@ -525,7 +535,6 @@ typedef struct exp_node {
 #define func_node    sub.nodep.x.extra
 #define prev_frame_size	sub.nodep.reflags
 #define reti         sub.nodep.l.li
-#define num_tail_calls    sub.nodep.cnt
 
 /* Node_var: */
 #define var_value    lnode
@@ -543,29 +552,16 @@ typedef struct exp_node {
 #define xarray		sub.nodep.rn
 #define parent_array	sub.nodep.x.extra
 
-#define ainit		array_funcs[0]
-#define ainit_ind	0
-#define atypeof		array_funcs[1]
-#define atypeof_ind	1
-#define alength		array_funcs[2]
-#define alength_ind	2
-#define alookup 	array_funcs[3]
-#define alookup_ind	3
-#define aexists 	array_funcs[4]
-#define aexists_ind	4
-#define aclear		array_funcs[5]
-#define aclear_ind	5
-#define aremove		array_funcs[6]
-#define aremove_ind	6
-#define alist		array_funcs[7]
-#define alist_ind	7
-#define acopy		array_funcs[8]
-#define acopy_ind	8
-#define adump		array_funcs[9]
-#define adump_ind	9
-#define astore		array_funcs[10]
-#define astore_ind	10
-#define NUM_AFUNCS	11		/* # of entries in array_funcs */
+#define ainit		array_funcs->init
+#define atypeof		array_funcs->type_of
+#define alookup 	array_funcs->lookup
+#define aexists 	array_funcs->exists
+#define aclear		array_funcs->clear
+#define aremove		array_funcs->remove
+#define alist		array_funcs->list
+#define acopy		array_funcs->copy
+#define adump		array_funcs->dump
+#define astore		array_funcs->store
 
 /* Node_array_ref: */
 #define orig_array lnode
@@ -576,9 +572,7 @@ typedef struct exp_node {
 #define alevel     sub.nodep.x.xl
 
 /* Op_comment	*/
-#define comment_type	sub.val.idx
-#define EOL_COMMENT 1
-#define FULL_COMMENT 2
+#define comment_type	sub.val.comtype
 
 /* --------------------------------lint warning types----------------------------*/
 typedef enum lintvals {
@@ -669,11 +663,13 @@ typedef enum opcodeval {
 	Op_K_next,
 	Op_K_exit,
 	Op_K_return,
+	Op_K_return_from_eval,
 	Op_K_delete,
 	Op_K_delete_loop,
 	Op_K_getline_redir,
 	Op_K_getline,
 	Op_K_nextfile,
+	Op_K_namespace,
 
 	Op_builtin,
 	Op_sub_builtin,		/* sub, gsub and gensub */
@@ -718,6 +714,7 @@ typedef enum opcodeval {
 	Op_exec_count,
 	Op_breakpoint,
 	Op_lint,
+	Op_lint_plus,
 	Op_atexit,
 	Op_stop,
 
@@ -753,6 +750,16 @@ enum redirval {
 
 struct break_point;
 
+#if __DECC && __VAX
+typedef unsigned long exec_count_t;	// for exec_count
+#define EXEC_COUNT_FMT	"%lu"
+#define EXEC_COUNT_PROFILE_FMT	"%6lu"
+#else
+typedef unsigned long long exec_count_t;	// for exec_count
+#define EXEC_COUNT_FMT	"%llu"
+#define EXEC_COUNT_PROFILE_FMT	"%6llu"
+#endif
+
 typedef struct exp_instruction {
 	struct exp_instruction *nexti;
 	union {
@@ -763,6 +770,7 @@ typedef struct exp_instruction {
 					awk_value_t *result,
 					struct awk_ext_func *finfo);
 		long dl;
+		exec_count_t ldl;	// for exec_count
 		char *name;
 	} d;
 
@@ -775,6 +783,7 @@ typedef struct exp_instruction {
 		awk_ext_func_t *exf;
 	} x;
 
+	struct exp_instruction *comment;
 	short source_line;
 	short pool_size;	// memory management in symbol.c
 	OPCODE opcode;
@@ -860,9 +869,6 @@ typedef struct exp_instruction {
 /* Op_func_call, Op_func */
 #define func_body       x.xn
 
-/* Op_func_call */
-#define tail_call	d.dl
-
 /* Op_subscript */
 #define sub_count       d.dl
 
@@ -905,7 +911,7 @@ typedef struct exp_instruction {
 
 /*------------------ pretty printing/profiling --------*/
 /* Op_exec_count */
-#define exec_count      d.dl
+#define exec_count      d.ldl
 
 /* Op_K_while */
 #define while_body      d.di
@@ -928,6 +934,9 @@ typedef struct exp_instruction {
 #define condpair_left   d.di
 #define condpair_right  x.xi
 
+/* Op_Rule, Op_Func */
+#define ns_name		d.name
+
 /* Op_store_var */
 #define initval         x.xn
 
@@ -946,30 +955,33 @@ typedef struct iobuf {
 	bool valid;
 	int errcode;
 
-	int flag;
-#		define	IOP_IS_TTY	1
-#		define  IOP_AT_EOF      2
-#		define  IOP_CLOSED      4
-#		define  IOP_AT_START    8
+	enum iobuf_flags {
+		IOP_IS_TTY	= 1,
+		IOP_AT_EOF	= 2,
+		IOP_CLOSED	= 4,
+		IOP_AT_START	= 8,
+	} flag;
 } IOBUF;
 
 typedef void (*Func_ptr)(void);
 
 /* structure used to dynamically maintain a linked-list of open files/pipes */
 struct redirect {
-	unsigned int flag;
-#		define	RED_FILE	1
-#		define	RED_PIPE	2
-#		define	RED_READ	4
-#		define	RED_WRITE	8
-#		define	RED_APPEND	16
-#		define	RED_NOBUF	32
-#		define	RED_USED	64	/* closed temporarily to reuse fd */
-#		define	RED_EOF		128
-#		define	RED_TWOWAY	256
-#		define	RED_PTY		512
-#		define	RED_SOCKET	1024
-#		define	RED_TCP		2048
+	enum redirect_flags {
+		RED_NONE	= 0,
+		RED_FILE	= 1,
+		RED_PIPE	= 2,
+		RED_READ	= 4,
+		RED_WRITE	= 8,
+		RED_APPEND	= 16,
+		RED_NOBUF	= 32,
+		RED_USED	= 64,	/* closed temporarily to reuse fd */
+		RED_EOF		= 128,
+		RED_TWOWAY	= 256,
+		RED_PTY		= 512,
+		RED_SOCKET	= 1024,
+		RED_TCP		= 2048,
+	} flag;
 	char *value;
 	FILE *ifp;	/* input fp, needed for PIPES_SIMULATED */
 	IOBUF *iop;
@@ -980,14 +992,15 @@ struct redirect {
 	const char *mode;
 	awk_output_buf_t output;
 };
+typedef enum redirect_flags redirect_flags_t;
 
 /* values for BINMODE, used as bit flags */
 
 enum binmode_values {
-	TEXT_TRANSLATE = 0,	/* usual \r\n ---> \n translation */
-	BINMODE_INPUT = 1,	/* no translation for input files */
-	BINMODE_OUTPUT = 2,	/* no translation for output files */
-	BINMODE_BOTH = 3	/* no translation for either */
+	TEXT_TRANSLATE	= 0,	/* usual \r\n ---> \n translation */
+	BINMODE_INPUT	= 1,	/* no translation for input files */
+	BINMODE_OUTPUT	= 2,	/* no translation for output files */
+	BINMODE_BOTH	= 3	/* no translation for either */
 };
 
 /*
@@ -1023,11 +1036,13 @@ typedef struct srcfile {
 	char *lexeme;
 	char *lexptr_begin;
 	int lasttok;
+	INSTRUCTION *comment;	/* comment on @load line */
+	const char *namespace;
 } SRCFILE;
 
 // structure for INSTRUCTION pool, needed mainly for debugger
 typedef struct instruction_pool {
-#define MAX_INSTRUCTION_ALLOC	3	// we don't call bcalloc with more than this
+#define MAX_INSTRUCTION_ALLOC	4	// we don't call bcalloc with more than this
 	struct instruction_mem_pool {
 		struct instruction_block *block_list;
 		INSTRUCTION *free_space;	// free location in active block
@@ -1061,13 +1076,16 @@ struct block_item {
 struct block_header {
 	struct block_item *freep;
 	size_t size;
+	const char *name;
+	long highwater;
+#ifdef MEMDEBUG
+	long active;
+#endif
 };
 
 enum block_id {
 	BLOCK_NODE = 0,
 	BLOCK_BUCKET,
-	BLOCK_MPFR,
-	BLOCK_MPZ,
 	BLOCK_MAX	/* count */
 };
 
@@ -1098,9 +1116,12 @@ extern int OFSlen;
 extern char *ORS;
 extern int ORSlen;
 extern char *OFMT;
-extern char *CONVFMT;
+extern const char *CONVFMT;
 extern int CONVFMTidx;
 extern int OFMTidx;
+#ifdef HAVE_MPFR
+extern int MPFR_round_mode;
+#endif
 extern char *TEXTDOMAIN;
 extern NODE *BINMODE_node, *CONVFMT_node, *FIELDWIDTHS_node, *FILENAME_node;
 extern NODE *FNR_node, *FS_node, *IGNORECASE_node, *NF_node;
@@ -1113,6 +1134,7 @@ extern NODE *Null_field;
 extern NODE **fields_arr;
 extern int sourceline;
 extern char *source;
+extern int errcount;
 extern int (*interpret)(INSTRUCTION *);	/* interpreter routine */
 extern NODE *(*make_number)(double);	/* double instead of AWKNUM on purpose */
 extern NODE *(*str2number)(NODE *);
@@ -1120,9 +1142,9 @@ extern NODE *(*format_val)(const char *, int, NODE *);
 extern int (*cmp_numbers)(const NODE *, const NODE *);
 
 /* built-in array types */
-extern afunc_t str_array_func[];
-extern afunc_t cint_array_func[];
-extern afunc_t int_array_func[];
+extern const array_funcs_t str_array_func;
+extern const array_funcs_t cint_array_func;
+extern const array_funcs_t int_array_func;
 
 /* special node used to indicate success in array routines (not NULL) */
 extern NODE *success_node;
@@ -1130,27 +1152,29 @@ extern NODE *success_node;
 extern struct block_header nextfree[];
 extern bool field0_valid;
 
-extern int do_flags;
+extern bool do_itrace;	/* separate so can poke from a debugger */
 
 extern SRCFILE *srcfiles; /* source files */
 
-enum do_flag_values {
-	DO_LINT_INVALID	= 0x0001,	/* only warn about invalid */
-	DO_LINT_ALL	= 0x0002,	/* warn about all things */
-	DO_LINT_OLD	= 0x0004,	/* warn about stuff not in V7 awk */
-	DO_TRADITIONAL	= 0x0008,	/* no gnu extensions, add traditional weirdnesses */
-	DO_POSIX	= 0x0010,	/* turn off gnu and unix extensions */
-	DO_INTL		= 0x0020,	/* dump locale-izable strings to stdout */
-	DO_NON_DEC_DATA	= 0x0040,	/* allow octal/hex C style DATA. Use with caution! */
-	DO_INTERVALS	= 0x0080,	/* allow {...,...} in regexps, see resetup() */
-	DO_PRETTY_PRINT	= 0x0100,	/* pretty print the program */
-	DO_DUMP_VARS	= 0x0200,	/* dump all global variables at end */
-	DO_TIDY_MEM	= 0x0400,	/* release vars when done */
-	DO_SANDBOX	= 0x0800,	/* sandbox mode - disable 'system' function & redirections */
-	DO_PROFILE	= 0x1000,	/* profile the program */
-	DO_DEBUG	= 0x2000,	/* debug the program */
-	DO_MPFR		= 0x4000	/* arbitrary-precision floating-point math */
-};
+extern enum do_flag_values {
+	DO_FLAG_NONE       = 0x00000,
+	DO_LINT_INVALID	   = 0x00001,	/* only warn about invalid */
+	DO_LINT_EXTENSIONS = 0x00002,	/* warn about gawk extensions */
+	DO_LINT_ALL	   = 0x00004,	/* warn about all things */
+	DO_LINT_OLD	   = 0x00008,	/* warn about stuff not in V7 awk */
+	DO_TRADITIONAL	   = 0x00010,	/* no gnu extensions, add traditional weirdnesses */
+	DO_POSIX	   = 0x00020,	/* turn off gnu and unix extensions */
+	DO_INTL		   = 0x00040,	/* dump locale-izable strings to stdout */
+	DO_NON_DEC_DATA	   = 0x00080,	/* allow octal/hex C style DATA. Use with caution! */
+	DO_INTERVALS	   = 0x00100,	/* allow {...,...} in regexps, see resetup() */
+	DO_PRETTY_PRINT	   = 0x00200,	/* pretty print the program */
+	DO_DUMP_VARS	   = 0x00400,	/* dump all global variables at end */
+	DO_TIDY_MEM	   = 0x00800,	/* release vars when done */
+	DO_SANDBOX	   = 0x01000,	/* sandbox mode - disable 'system' function & redirections */
+	DO_PROFILE	   = 0x02000,	/* profile the program */
+	DO_DEBUG	   = 0x04000,	/* debug the program */
+	DO_MPFR		   = 0x08000,	/* arbitrary-precision floating-point math */
+} do_flags;
 
 #define do_traditional      (do_flags & DO_TRADITIONAL)
 #define do_posix            (do_flags & DO_POSIX)
@@ -1172,9 +1196,11 @@ extern int exit_val;
 #ifdef NO_LINT
 #define do_lint 0
 #define do_lint_old 0
+#define do_lint_extensions 0
 #else
 #define do_lint             (do_flags & (DO_LINT_INVALID|DO_LINT_ALL))
 #define do_lint_old         (do_flags & DO_LINT_OLD)
+#define do_lint_extensions  (do_flags & DO_LINT_EXTENSIONS)
 #endif
 extern int gawk_mb_cur_max;
 
@@ -1200,13 +1226,17 @@ extern bool do_ieee_fmt;	/* emulate IEEE 754 floating-point format */
 extern const char *myname;
 extern const char def_strftime_format[];
 
-extern char quote;
-extern char *defpath;
-extern char *deflibpath;
-extern char envsep;
+extern const char quote;
+extern const char *defpath;
+extern const char *deflibpath;
+extern const char envsep;
 
 extern char casetable[];	/* for case-independent regexp matching */
 
+extern const char awk_namespace[];	/* "awk" */
+extern const char *current_namespace;
+extern bool namespace_changed;
+
 /* ------------------------- Runtime stack -------------------------------- */
 
 typedef union stack_item {
@@ -1251,8 +1281,11 @@ static inline void
 DEREF(NODE *r)
 {
 	assert(r->valref > 0);
-	if (--r->valref == 0)
-		r_unref(r);
+#ifndef GAWKDEBUG
+	if (--r->valref > 0)
+		return;
+#endif
+	r_unref(r);
 }
 
 #define POP_NUMBER() force_number(POP_SCALAR())
@@ -1287,7 +1320,7 @@ DEREF(NODE *r)
 
 #define get_number_uj(n)	numtype_choose((n), mpfr_get_uj((n)->mpg_numbr, ROUND_MODE), (uintmax_t) mpz_get_d((n)->mpg_i), (uintmax_t) (n)->numbr)
 
-#define iszero(n)		numtype_choose((n), mpfr_zero_p((n)->mpg_numbr), (mpz_sgn((n)->mpg_i) == 0), ((n)->numbr == 0.0))
+#define is_zero(n)		numtype_choose((n), mpfr_zero_p((n)->mpg_numbr), (mpz_sgn((n)->mpg_i) == 0), ((n)->numbr == 0.0))
 
 #define IEEE_FMT(r, t)		(void) (do_ieee_fmt && format_ieee(r, t))
 
@@ -1305,7 +1338,7 @@ DEREF(NODE *r)
 #define is_mpg_number(n)	0
 #define is_mpg_float(n)		0
 #define is_mpg_integer(n)	0
-#define iszero(n)		((n)->numbr == 0.0)
+#define is_zero(n)		((n)->numbr == 0.0)
 #endif
 
 #define var_uninitialized(n)	((n)->var_value == Nnull_string)
@@ -1313,12 +1346,23 @@ DEREF(NODE *r)
 #define get_lhs(n, r)	 (n)->type == Node_var && ! var_uninitialized(n) ? \
 				&((n)->var_value) : r_get_lhs((n), (r))
 
+#ifdef MEMDEBUG
+
+extern void *r_getblock(int id);
+extern void r_freeblock(void *, int id);
+#define getblock(p, id, ty)	(void) (p = (ty) r_getblock(id))
+#define freeblock(p, id)	(void) (r_freeblock(p, id))
+
+#else /* MEMDEBUG */
+
 #define getblock(p, id, ty)  (void) ((p = (ty) nextfree[id].freep) ? \
 			(ty) (nextfree[id].freep = ((struct block_item *) p)->freep) \
 			: (p = (ty) more_blocks(id)))
 #define freeblock(p, id)	 (void) (((struct block_item *) p)->freep = nextfree[id].freep, \
 					nextfree[id].freep = (struct block_item *) p)
 
+#endif /* MEMDEBUG */
+
 #define getnode(n)	getblock(n, BLOCK_NODE, NODE *)
 #define freenode(n)	freeblock(n, BLOCK_NODE)
 
@@ -1327,8 +1371,10 @@ DEREF(NODE *r)
 
 #define	make_string(s, l)	make_str_node((s), (l), 0)
 
+// Flags for making string nodes
 #define		SCAN			1
 #define		ALREADY_MALLOCED	2
+#define		ELIDE_BACK_NL		4
 
 #define	cant_happen()	r_fatal("internal error line %d, file: %s", \
 				__LINE__, __FILE__)
@@ -1344,14 +1390,7 @@ DEREF(NODE *r)
 extern jmp_buf fatal_tag;
 extern int fatal_tag_valid;
 
-#define PUSH_BINDING(stack, tag, val)	\
-if (val++) \
-	memcpy((char *) (stack), (const char *) tag, sizeof(jmp_buf))
-#define POP_BINDING(stack, tag, val)	\
-if (--val) \
-	memcpy((char *) tag, (const char *) (stack), sizeof(jmp_buf))
-
-#define assoc_length(a)	((*((a)->alength(a, NULL)))->table_size)
+#define assoc_length(a)	((a)->table_size)
 #define assoc_empty(a)	(assoc_length(a) == 0)
 #define assoc_lookup(a, s)	((a)->alookup(a, s))
 
@@ -1361,6 +1400,7 @@ if (--val) \
 /* assoc_remove --- remove an index from symbol[] */
 #define assoc_remove(a, s) ((a)->aremove(a, s) != NULL)
 
+
 /* ------------- Function prototypes or defs (as appropriate) ------------- */
 /* array.c */
 typedef enum { SORTED_IN = 1, ASORT, ASORTI } sort_context_t;
@@ -1383,8 +1423,6 @@ extern NODE *force_array(NODE *symbol, bool canfatal);
 extern const char *make_aname(const NODE *symbol);
 extern const char *array_vname(const NODE *symbol);
 extern void array_init(void);
-extern int register_array_func(afunc_t *afunc);
-extern NODE **null_length(NODE *symbol, NODE *subs);
 extern NODE **null_afunc(NODE *symbol, NODE *subs);
 extern void set_SUBSEP(void);
 extern NODE *concat_exp(int nargs, bool do_subsep);
@@ -1400,13 +1438,14 @@ extern NODE *do_asort(int nargs);
 extern NODE *do_asorti(int nargs);
 extern unsigned long (*hash)(const char *s, size_t len, unsigned long hsize, size_t *code);
 extern void init_env_array(NODE *env_node);
+extern void init_argv_array(NODE *argv_node, NODE *shadow_node);
 /* awkgram.c */
 extern NODE *variable(int location, char *name, NODETYPE type);
-extern int parse_program(INSTRUCTION **pcode);
+extern int parse_program(INSTRUCTION **pcode, bool from_eval);
 extern void track_ext_func(const char *name);
 extern void dump_funcs(void);
 extern void dump_vars(const char *fname);
-extern const char *getfname(NODE *(*)(int));
+extern const char *getfname(NODE *(*)(int), bool prepend_awk);
 extern NODE *stopme(int nargs);
 extern void shadow_funcs(void);
 extern int check_special(const char *name);
@@ -1422,7 +1461,8 @@ extern bool is_alpha(int c);
 extern bool is_alnum(int c);
 extern bool is_letter(int c);
 extern bool is_identchar(int c);
-extern NODE *make_regnode(int type, NODE *exp);
+extern NODE *make_regnode(NODETYPE type, NODE *exp);
+extern bool validate_qualified_name(char *token);
 /* builtin.c */
 extern double double_to_int(double d);
 extern NODE *do_exp(int nargs);
@@ -1472,6 +1512,10 @@ extern NODE *do_typeof(int nargs);
 extern int strncasecmpmbs(const unsigned char *,
 			  const unsigned char *, size_t);
 extern int sanitize_exit_status(int status);
+extern void check_symtab_functab(NODE *dest, const char *fname, const char *msg);
+/* debug.c */
+extern void init_debug(void);
+extern int debug_prog(INSTRUCTION *pc);
 /* eval.c */
 extern void PUSH_CODE(INSTRUCTION *cp);
 extern INSTRUCTION *POP_CODE(void);
@@ -1509,8 +1553,9 @@ extern NODE **r_get_field(NODE *n, Func_ptr *assign, bool reference);
 extern NODE *do_ext(int nargs);
 void load_ext(const char *lib_name);	/* temporary */
 extern void close_extensions(void);
+extern bool is_valid_identifier(const char *name);
 #ifdef DYNAMIC
-extern awk_bool_t make_builtin(const awk_ext_func_t *);
+extern awk_bool_t make_builtin(const char *name_space, const awk_ext_func_t *);
 extern NODE *get_argument(int);
 extern NODE *get_actual_argument(NODE *, int, bool);
 #define get_scalar_argument(n, i)  get_actual_argument((n), (i), false)
@@ -1551,7 +1596,7 @@ extern void print_ext_versions(void);
 extern void free_api_string_copies(void);
 
 /* gawkmisc.c */
-extern char *gawk_name(const char *filespec);
+extern const char *gawk_name(const char *filespec);
 extern void os_arg_fixup(int *argcp, char ***argvp);
 extern int os_devopen(const char *name, int flag);
 extern void os_close_on_exec(int fd, const char *name, const char *what, const char *dir);
@@ -1580,7 +1625,7 @@ extern struct redirect *redirect_string(const char *redir_exp_str,
 		int *errflg, int extfd, bool failure_fatal);
 extern NODE *do_close(int nargs);
 extern int flush_io(void);
-extern int close_io(bool *stdio_problem);
+extern int close_io(bool *stdio_problem, bool *got_EPIPE);
 typedef enum { CLOSE_ALL, CLOSE_TO, CLOSE_FROM } two_way_close_type;
 extern int close_rp(struct redirect *rp, two_way_close_type how);
 extern int devopen_simple(const char *name, const char *mode, bool try_real_open);
@@ -1606,6 +1651,7 @@ extern char *estrdup(const char *str, size_t len);
 extern void update_global_values();
 extern long getenv_long(const char *name);
 extern void after_beginfile(IOBUF **curfile);
+extern void set_current_namespace(const char *new_namespace);
 
 /* mpfr.c */
 extern void set_PREC(void);
@@ -1638,6 +1684,7 @@ extern void cleanup_mpfr(void);
 extern NODE *mpg_node(unsigned int);
 extern const char *mpg_fmt(const char *, ...);
 extern int mpg_strtoui(mpz_ptr, char *, size_t, char **, int);
+extern void mpg_zero(NODE *n);
 #endif
 /* msg.c */
 extern void gawk_exit(int status);
@@ -1684,6 +1731,9 @@ extern wint_t btowc_cache[];
 #define btowc_cache(x) btowc_cache[(x)&0xFF]
 extern void init_btowc_cache();
 #define is_valid_character(b)	(btowc_cache[(b)&0xFF] != WEOF)
+extern bool out_of_range(NODE *n);
+extern char *format_nan_inf(NODE *n, char format);
+extern bool is_ieee_magic_val(const char *val);
 /* re.c */
 extern Regexp *make_regexp(const char *s, size_t len, bool ignorecase, bool dfa, bool canfatal);
 extern int research(Regexp *rp, char *str, int start, size_t len, int flags);
@@ -1723,6 +1773,7 @@ extern NODE **variable_list();
 extern NODE **function_list(bool sort);
 extern void print_vars(NODE **table, Func_print print_func, FILE *fp);
 extern bool check_param_names(void);
+extern bool is_all_upper(const char *name);
 
 /* floatcomp.c */
 #ifdef HAVE_UINTMAX_T
@@ -1772,9 +1823,15 @@ extern uintmax_t adjust_uint(uintmax_t n);
 /* POP_ARRAY --- get the array at the top of the stack */
 
 static inline NODE *
-POP_ARRAY()
+POP_ARRAY(bool check_for_untyped)
 {
 	NODE *t = POP();
+	static bool warned = false;
+
+	if (do_lint && ! warned && check_for_untyped && t->type == Node_var_new) {
+		warned = true;
+		lintwarn(_("behavior of `for' loop on untyped variable is not defined by POSIX"));
+	}
 
 	return (t->type == Node_var_array) ? t : force_array(t, true);
 }
@@ -1856,14 +1913,20 @@ dupnode(NODE *n)
  * and OFMT values. But if the value entered gawk as a string or strnum, then
  * stfmt should be set to STFMT_UNUSED, and the string representation should
  * not change.
+ *
+ * Additional twist: If ROUNDMODE changed at some point we have to
+ * recompute also.
  */
 
 static inline NODE *
 force_string_fmt(NODE *s, const char *fmtstr, int fmtidx)
 {
 	if ((s->flags & STRCUR) != 0
-		    && (s->stfmt == STFMT_UNUSED || s->stfmt == fmtidx)
-	)
+		&& (s->stfmt == STFMT_UNUSED || (s->stfmt == fmtidx
+#ifdef HAVE_MPFR
+						&& s->strndmode == MPFR_round_mode
+#endif
+				)))
 		return s;
 	return format_val(fmtstr, fmtidx, s);
 }
@@ -1883,6 +1946,7 @@ force_string_fmt(NODE *s, const char *fmtstr, int fmtidx)
 static inline void
 unref(NODE *r)
 {
+	assert(r == NULL || r->valref > 0);
 	if (r != NULL && --r->valref <= 0)
 		r_unref(r);
 }
@@ -1933,7 +1997,7 @@ boolval(NODE *t)
 {
 	(void) fixtype(t);
 	if ((t->flags & NUMBER) != 0)
-		return ! iszero(t);
+		return ! is_zero(t);
 	return (t->stlen > 0);
 }
 
@@ -1949,7 +2013,7 @@ emalloc_real(size_t count, const char *where, const char *var, const char *file,
 
 	ret = (void *) malloc(count);
 	if (ret == NULL)
-		fatal(_("%s:%d:%s: %s: can't allocate %ld bytes of memory (%s)"),
+		fatal(_("%s:%d:%s: %s: cannot allocate %ld bytes of memory: %s"),
 			file, line, where, var, (long) count, strerror(errno));
 
 	return ret;
@@ -1967,7 +2031,7 @@ ezalloc_real(size_t count, const char *where, const char *var, const char *file,
 
 	ret = (void *) calloc(1, count);
 	if (ret == NULL)
-		fatal(_("%s:%d:%s: %s: can't allocate %ld bytes of memory (%s)"),
+		fatal(_("%s:%d:%s: %s: cannot allocate %ld bytes of memory: %s"),
 			file, line, where, var, (long) count, strerror(errno));
 
 	return ret;
@@ -1985,7 +2049,7 @@ erealloc_real(void *ptr, size_t count, const char *where, const char *var, const
 
 	ret = (void *) realloc(ptr, count);
 	if (ret == NULL)
-		fatal(_("%s:%d:%s: %s: can't reallocate %ld bytes of memory (%s)"),
+		fatal(_("%s:%d:%s: %s: cannot reallocate %ld bytes of memory: %s"),
 			file, line, where, var, (long) count, strerror(errno));
 
 	return ret;
@@ -2005,6 +2069,20 @@ make_number_node(unsigned int flags)
 	return r;
 }
 
+/* assoc_set -- set an element in an array. Does unref(sub)! */
+
+static inline void
+assoc_set(NODE *array, NODE *sub, NODE *value)
+{
+
+	NODE **lhs = assoc_lookup(array, sub);
+	unref(*lhs);
+	*lhs = value;
+	if (array->astore != NULL)
+		(*array->astore)(array, sub);
+	unref(sub);
+}
+
 /*
  * str_terminate_f, str_terminate, str_restore: function and macros to
  * reduce chances of typos when terminating and restoring strings.
