@@ -3,7 +3,7 @@
  */
 
 /*
- * Copyright (C) 1991-2018 the Free Software Foundation, Inc.
+ * Copyright (C) 1991-2019, 2021 the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -107,7 +107,9 @@ make_regexp(const char *s, size_t len, bool ignorecase, bool dfa, bool canfatal)
 		}
 
 		const char *ok_to_escape;
-		if (do_traditional)
+		if (do_posix)
+			ok_to_escape = "{}()|*+?.^$\\[]/-";
+		else if (do_traditional)
 			ok_to_escape = "()|*+?.^$\\[]/-";
 		else
 			ok_to_escape = "<>`'BywWsS{}()|*+?.^$\\[]/-";
@@ -256,10 +258,10 @@ make_regexp(const char *s, size_t len, bool ignorecase, bool dfa, bool canfatal)
 		refree(rp);
 		if (! canfatal) {
 			/* rerr already gettextized inside regex routines */
-			error("%s: /%s/", rerr, buf);
+			error("%s: /%.*s/", rerr, (int) len, s);
  			return NULL;
 		}
-		fatal("invalid regexp: %s: /%s/", rerr, buf);
+		fatal("invalid regexp: %s: /%.*s/", rerr, (int) len, s);
 	}
 
 	/* gack. this must be done *after* re_compile_pattern */
@@ -280,7 +282,7 @@ make_regexp(const char *s, size_t len, bool ignorecase, bool dfa, bool canfatal)
 	}
 
 	for (i = len - 1; i >= 0; i--) {
-		if (strchr("*+|?", buf[i]) != NULL) {
+		if (strchr("*+|?{}", buf[i]) != NULL) {
 			rp->maybe_long = true;
 			break;
 		}
@@ -389,6 +391,16 @@ dfaerror(const char *s)
 	exit(EXIT_FATAL);	/* for DJGPP */
 }
 
+/* re_cache_get --- populate regexp cache if empty */
+
+static inline Regexp *
+re_cache_get(NODE *t)
+{
+	if (t->re_reg[IGNORECASE] == NULL)
+		t->re_reg[IGNORECASE] = make_regexp(t->re_exp->stptr, t->re_exp->stlen, IGNORECASE, t->re_cnt, true);
+	return t->re_reg[IGNORECASE];
+}
+
 /* re_update --- recompile a dynamic regexp */
 
 Regexp *
@@ -397,18 +409,18 @@ re_update(NODE *t)
 	NODE *t1;
 
 	if (t->type == Node_val && (t->flags & REGEX) != 0)
-		return t->typed_re->re_reg[IGNORECASE];
+		return re_cache_get(t->typed_re);
 
 	if ((t->re_flags & CONSTANT) != 0) {
 		/* it's a constant, so just return it as is */
 		assert(t->type == Node_regex);
-		return t->re_reg[IGNORECASE];
+		return re_cache_get(t);
 	}
 	t1 = t->re_exp;
 	if (t->re_text != NULL) {
 		/* if contents haven't changed, just return it */
 		if (cmp_nodes(t->re_text, t1, true) == 0)
-			return t->re_reg[IGNORECASE];
+			return re_cache_get(t);
 		/* things changed, fall through to recompile */
 		unref(t->re_text);
 	}
@@ -418,13 +430,20 @@ re_update(NODE *t)
 	/* text changed */
 
 	/* free old */
-	if (t->re_reg[0] != NULL)
+	if (t->re_reg[0] != NULL) {
 		refree(t->re_reg[0]);
-	if (t->re_reg[1] != NULL)
+		t->re_reg[0] = NULL;
+	}
+	if (t->re_reg[1] != NULL) {
 		refree(t->re_reg[1]);
-	if (t->re_cnt > 0)
-		t->re_cnt++;
-	if (t->re_cnt > 10)
+		t->re_reg[1] = NULL;
+	}
+	if (t->re_cnt > 0 && ++t->re_cnt > 10)
+		/*
+		 * The regex appears to update frequently, so disable DFA
+		 * matching (which trades off expensive upfront compilation
+		 * overhead for faster subsequent matching).
+		 */
 		t->re_cnt = 0;
 	if (t->re_text == NULL) {
 		/* reset regexp text if needed */
@@ -432,13 +451,7 @@ re_update(NODE *t)
 		unref(t->re_text);
 		t->re_text = dupnode(t1);
 	}
-	/* compile it */
-	t->re_reg[0] = make_regexp(t->re_text->stptr, t->re_text->stlen,
-				false, t->re_cnt, true);
-	t->re_reg[1] = make_regexp(t->re_text->stptr, t->re_text->stlen,
-				true, t->re_cnt, true);
-
-	return t->re_reg[IGNORECASE];
+	return re_cache_get(t);
 }
 
 /* resetup --- choose what kind of regexps we match */
@@ -600,7 +613,7 @@ check_bracket_exp(char *s, size_t length)
 	sp = s;
 
 again:
-	sp = sp2 = memchr(sp, '[', (end - sp));
+	sp = sp2 = (char *) memchr(sp, '[', (end - sp));
 	if (sp == NULL)
 		goto done;
 
