@@ -3,7 +3,7 @@
  */
 
 /*
- * Copyright (C) 1986, 1988, 1989, 1991-2015, 2017, 2018, 2019,
+ * Copyright (C) 1986, 1988, 1989, 1991-2015, 2017-2020,
  * the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
@@ -39,6 +39,7 @@ static void (*install_func)(NODE *) = NULL;
 static NODE *make_symbol(const char *name, NODETYPE type);
 static NODE *install(const char *name, NODE *parm, NODETYPE type);
 static void free_bcpool(INSTRUCTION_POOL *pl);
+static NODE *get_name_from_awk_ns(const char *name);
 
 static AWK_CONTEXT *curr_ctxt = NULL;
 static int ctxt_level;
@@ -102,10 +103,7 @@ lookup(const char *name)
 	tables[3] = symbol_table;	/* then globals */
 	tables[4] = NULL;
 
-	if (strncmp(name, "awk::", 5) == 0)
-		tmp = make_string(name + 5, strlen(name) - 5);
-	else
-		tmp = make_string(name, strlen(name));
+	tmp = get_name_from_awk_ns(name);
 
 	n = NULL;
 	for (i = 0; tables[i] != NULL; i++) {
@@ -292,6 +290,7 @@ make_symbol(const char *name, NODETYPE type)
 		r->var_value = dupnode(Nnull_string);
 	r->vname = (char *) name;
 	r->type = type;
+	r->valref = 1;
 
 	return r;
 }
@@ -306,7 +305,7 @@ install(const char *name, NODE *parm, NODETYPE type)
 	NODE *n_name;
 	NODE *prev;
 
-	n_name = make_string(name, strlen(name));
+	n_name = get_name_from_awk_ns(name);
 
 	table = symbol_table;
 
@@ -363,7 +362,18 @@ comp_symbol(const void *v1, const void *v2)
 	n1 = *npp1;
 	n2 = *npp2;
 
-	return strcmp(n1->vname, n2->vname);
+	// names in awk namespace come out first
+	bool n1_is_in_ns = (strchr(n1->vname, ':') != NULL);
+	bool n2_is_in_ns = (strchr(n2->vname, ':') != NULL);
+
+	if (n1_is_in_ns && n2_is_in_ns)
+		return strcmp(n1->vname, n2->vname);
+	else if (n1_is_in_ns && ! n2_is_in_ns)
+		return 1;
+	else if (! n1_is_in_ns && n2_is_in_ns)
+		return -1;
+	else
+		return strcmp(n1->vname, n2->vname);
 }
 
 
@@ -685,7 +695,7 @@ check_param_names(void)
 
 			if (in_array(func_table, & n)) {
 				error(
-			_("function `%s': can't use function `%s' as a parameter name"),
+			_("function `%s': cannot use function `%s' as a parameter name"),
 					list[i]->stptr,
 					f->fparms[j].param);
 				result = false;
@@ -815,7 +825,7 @@ pop_context()
 
 	assert(curr_ctxt != NULL);
 	if (curr_ctxt->prev == NULL)
-		fatal(_("can not pop main context"));
+		fatal(_("cannot pop main context"));
 	ctxt = curr_ctxt->prev;
 	/* restore source and sourceline */
 	sourceline = ctxt->sourceline;
@@ -973,3 +983,18 @@ is_all_upper(const char *name)
 
 	return true;
 }
+
+/* get_name_from_awk_ns --- get the name after awk:: or the full name */
+
+static NODE *
+get_name_from_awk_ns(const char *name)
+{
+	NODE *tmp;
+
+	if (strncmp(name, "awk::", 5) == 0)
+		tmp = make_string(name + 5, strlen(name) - 5);
+	else
+		tmp = make_string(name, strlen(name));
+
+	return tmp;
+}
